// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:rxdart/subjects.dart';
import 'package:wyatt_crud_bloc/src/core/enums/operation_type.dart';
import 'package:wyatt_crud_bloc/src/core/enums/where_query_type.dart';
import 'package:wyatt_crud_bloc/src/data/data_sources/crud_data_source.dart';
import 'package:wyatt_crud_bloc/src/domain/entities/query.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

extension _ModelExtension on Map<String, dynamic> {
  Map<String, dynamic> merge(
    Map<String, dynamic>? other, {
    bool skipNulls = false,
  }) {
    if (other == null) {
      return this;
    }
    final Map<String, dynamic> result = {};
    for (final key in keys) {
      result[key] = other[key] ?? this[key];
    }
    for (final key in other.keys) {
      if (!containsKey(key)) {
        result[key] = other[key];
      }
    }

    if (skipNulls) {
      result.removeWhere((key, value) => other[key] == null);
    }
    return result;
  }
}

/// {@template crud_data_source_in_memory_impl}
/// A [CrudDataSource] that stores data in memory.
/// {@endtemplate}
class CrudDataSourceInMemoryImpl extends CrudDataSource {
  /// {@macro crud_data_source_in_memory_impl}
  CrudDataSourceInMemoryImpl({
    required super.id,
    Map<String, Map<String, dynamic>>? data,
  }) : _data = data ?? {};

  final Map<String, Map<String, dynamic>> _data;
  final BehaviorSubject<List<Map<String, dynamic>?>> _streamData =
      BehaviorSubject<List<Map<String, dynamic>?>>.seeded([]);

  @override
  Future<void> create(Map<String, dynamic> object, {String? id}) async {
    _data[id ?? this.id(OperationType.create, object) ?? ''] = object;

    _streamData.add(_data.values.toList());
  }

  @override
  Future<void> delete(String id) async {
    _data.remove(id);
    _streamData.add(_data.values.toList());
  }

  @override
  Future<void> deleteAll() async {
    _data.clear();
    _streamData.add(_data.values.toList());
  }

  @override
  Future<Map<String, dynamic>?> get(String id) async => _data[id];

  @override
  Future<List<Map<String, dynamic>?>> getAll() async => _data.values.toList();

  @override
  Future<List<Map<String, dynamic>?>> search(List<Query> conditions) async {
    List<Map<String, dynamic>> result = _data.values.toList();

    for (final c in conditions) {
      if (c is WhereQuery) {
        result = result.where((m) => _whereQuery(c, m)).toList();
      }
      if (c is LimitQuery) {
        final limit = result.length - c.limit;
        result = result.sublist(limit >= 0 ? limit : 0);
      }
      if (c is OrderByQuery) {
        try {
          result.sort();
        } catch (_) {}
        if (c.ascending) {
          result = result.reversed.toList();
        }
      }
    }

    result.cast<Map<String, dynamic>>();

    return result;
  }

  @override
  Stream<List<Map<String, dynamic>?>> stream({
    String? id,
    List<Query>? conditions,
    bool includeMetadataChanges = false,
  }) =>
      _streamData.stream.map((result) {
        if (conditions == null) {
          return result;
        }

        List<Map<String, dynamic>?> res = result;

        for (final c in conditions) {
          if (c is WhereQuery) {
            res = res.where((element) => _whereQuery(c, element)).toList();
          }
          if (c is LimitQuery) {
            res = res.sublist(res.length - c.limit);
          }
          if (c is OrderByQuery) {
            res.sort();
            if (c.ascending) {
              res = res.reversed.toList();
            }
          }
        }

        return res;
      });

  @override
  Future<void> update(
    String id, {
    Map<String, dynamic>? object,
  }) {
    if (object != null && _data.containsKey(id)) {
      _data[id] = _data[id]!.merge(object);
      _streamData.add(_data.values.toList());
    }
    return Future.value();
  }

  @override
  Future<void> updateAll(Map<String, Object?>? data) {
    if (data != null) {
      for (final entry in _data.entries) {
        _data[entry.key] = entry.value.merge(data);
      }
      _streamData.add(_data.values.toList());
    }
    return Future.value();
  }

  bool _whereQuery(Query condition, Map<String, dynamic>? object) {
    if (object == null) {
      return false;
    }
    if (condition is WhereQuery) {
      switch (condition.type) {
        case WhereQueryType.isEqualTo:
          return object[condition.field] == condition.value;
        case WhereQueryType.isNotEqualTo:
          return object[condition.field] != condition.value;
        case WhereQueryType.isLessThan:
          return (object[condition.field] as num?) < (condition.value as num?);
        case WhereQueryType.isLessThanOrEqualTo:
          return (object[condition.field] as num?) <= (condition.value as num?);
        case WhereQueryType.isGreaterThan:
          return (object[condition.field] as num?) > (condition.value as num?);
        case WhereQueryType.isGreaterThanOrEqualTo:
          return (object[condition.field] as num?) >= (condition.value as num?);
        case WhereQueryType.arrayContains:
          return (object[condition.field] as List<Object>?)
                  ?.contains(condition.value) ??
              false;
        case WhereQueryType.arrayContainsAny:
          bool res = false;
          for (final o in condition.value as List<Object>) {
            res = (object[condition.field] as List<Object>?)?.contains(o) ??
                false;
          }
          return res;
        case WhereQueryType.whereIn:
          return (condition.value as List<Object>)
              .contains(object[condition.field]);
        case WhereQueryType.whereNotIn:
          return !(condition.value as List<Object>)
              .contains(object[condition.field]);
        case WhereQueryType.isNull:
          return object[condition.field] == null;
      }
    }
    return true;
  }
}
