// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:go_router/go_router.dart';
import 'package:wyatt_go_router/src/core/extensions/go_route_extension.dart';

extension GoRouteIterableExtension on Iterable<GoRoute> {
  /// Returns a list of public routes.
  List<GoRoute> get publicRoutes =>
      where((route) => route.isPublic).toList(growable: false);

  /// Returns a list of protected routes.
  List<GoRoute> get protectedRoutes =>
      where((route) => route.isProtected).toList(growable: false);

  /// Returns a list of routes that are neither public nor protected.
  List<GoRoute> get noneRoutes =>
      where((route) => route.isNone).toList(growable: false);

  /// Returns a go route from its name.
  ///
  /// If no route is found, a [StateError] is thrown.
  GoRoute fromName(String name) => firstWhere((route) => route.name == name);

  /// Returns a go route from its path.
  ///
  /// If no route is found, a [StateError] is thrown.
  GoRoute fromPath(String path) => firstWhere((route) => route.path == path);

  /// Returns a go route from its name or null if no route is found.
  GoRoute? fromNameOrNull(String name) {
    if (any((route) => route.name == name)) {
      return fromName(name);
    }

    return null;
  }

  /// Returns a go route from its path or null if no route is found.
  GoRoute? fromPathOrNull(String path) {
    if (any((route) => route.path == path)) {
      return fromPath(path);
    }

    return null;
  }

  /// Returns a go route from its name, or null if no route is found.
  ///
  /// This is a shorthand for [fromNameOrNull].
  GoRoute? operator [](String name) => fromNameOrNull(name);
}
